
/*-------------------------------------------------------------
  module: toesacQXshoupZPXMPI.C
  author: Angel L. Diaz

  date: 10/30/96

  description: 
     Interpolates in parallel a factor of an nXn symetric toeplitz 
     determinant utilizing 
        - SACLIB's rational field and polynomial arithmetic for 
          black box construction 
        - Shoup's modular arithmetic for black box evaluation
        - black box MPI interface for paralelization

  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.
 
-------------------------------------------------------------*/

#include "PlugIns/saclib.h"
#include "PlugIns/shoup.h"
#include "PlugIns/maps.h"
#include "BlackBox/CommonObjects/bbtoeplitz.h"
#include "BlackBox/Aux/bbselectormpi.h"
#include "BlackBox/Aux/bbvector.h"
#include "BlackBox/Algorithms/bbfactors.h"
#include "BlackBox/HomomorphicMaps/bbfachmapmpi.h"
#include "BlackBox/Algorithms/bbsparse.h"
#include "BlackBox/MPI/bbservermpi.h"
#include "BlackBox/Aux/bbtime.h"

typedef BlackBoxSymToeDet< SaclibQ, SaclibQX >            BBSymToeDetSacQ;
typedef BlackBoxSymToeDet< SaclibZP, SaclibZPX >          BBSymToeDetSacZP;
typedef BlackBoxFactors< SaclibQ, SaclibQX, 
        BBSymToeDetSacQ >                                 BBFactors;
typedef BlackBoxFactorsHMapMPI< SaclibQ, SaclibQX, 
        SaclibZP, SaclibZPX, BBSymToeDetSacQ, 
        BBSymToeDetSacZP, SaclibQSaclibZP >               BBFactorsHMapMPI;

typedef BlackBoxSymToeDet< ShoupZP, ShoupZPX >            BBSymToeDetShoupZP;
typedef BlackBoxFactorsHMapMPI< ShoupZP, ShoupZPX, 
        ShoupZP, ShoupZPX, BBSymToeDetShoupZP, 
        BBSymToeDetShoupZP, SaclibQShoupZP >         BBFactorsHMapShoupZPMPI;
typedef BlackBoxSelectorMPI< ShoupZP, 
        BBFactorsHMapShoupZPMPI >                         BBFactorShoupZPMPI;
typedef BlackBoxServerMPI< ShoupZP, ShoupZPX, 
        BBSymToeDetShoupZP >                              BBServerShoupZPXMPI;
 
int main( int argc, char *argv[] ) 
{
   MPI_Init(&argc, &argv);
   int    MyRank;                          // Rank of process 
   int    NumProc;                         // Number of processes 
   int    i,N, DegDet, DegVar,Fac,Seed,IsRestart;
   double Prob;
   char   ProbName[BBDEFS_FILEZ];
   char   FileName[BBDEFS_FILEZ];
   MP_INT MPCard;

   mpz_init_set_ui( &MPCard, 100000007 ); 
   MPI_Comm_rank(MPI_COMM_WORLD, &MyRank);
   MPI_Comm_size(MPI_COMM_WORLD, &NumProc);
   Prob      = 1.0;
   Seed      = 1;
   N         = 4;
   DegDet    = 4;
   DegVar    = 4;
   Fac       = 0;
   IsRestart = BBDEFS_FALSE;

   if ( MyRank == BBMPI_DRIVER_RANK ) {
      Word stack;
      SaclibInitEnv(1000000,stack);
      SaclibSetPrime( &MPCard );
      BBSymToeDetSacQ   TSacQ( N, DegDet , Prob );
      BBSymToeDetSacZP  TSacZP( N, DegDet , Prob );

      GetCPUTime( "toesacQXsacZPX.C" );

      //
      // construct utilizing SACLIB's rational arithmetic
      //

      BBFactors        FactorsTSacQ( TSacQ, Prob, Seed, &MPCard );
      SaclibQSaclibZP  h;
      BBFactorsHMapMPI FactorsTSacZPMPI( TSacZP, FactorsTSacQ, h );

      sprintf( FileName, "%dx%dtoesacQXshoupZPXMPI.ins", N, N );
      ofstream OutFile( FileName, ios::out );
      if ( !OutFile ) {
         cerr << "toefacshmap : cannot open " << FileName
              << " for output" << endl;
         exit(-1);
      }

      OutFile << FactorsTSacZPMPI;
      OutFile.close();
      SaclibCleanUpEnv();

      //
      // interpolate utilizing Shoup's modular arithmetic
      //

      ShoupInitEnv( &MPCard );

      BBSymToeDetShoupZP  TShoupZP( N, DegDet , Prob );
      BBFactorsHMapShoupZPMPI FactorsTShoupZPMPI( TShoupZP );

      ifstream InFile( FileName, ios::in );
      if ( !InFile ) {
         cerr << "toefacshmap : cannot open ``4x4toeZP.ins'' for input" << endl;
      }

      InFile >> FactorsTShoupZPMPI;
      InFile.close();

      BBFactorShoupZPMPI    FirstFacTShoupZPMPI( FactorsTShoupZPMPI, Fac );
      BBSPARSEAUX_VarVector Vars;
      BBSPARSEAUX_DegVector Degs;
      for( i=1; i<=N; i++ ) Vars.push_back(i);
      for( i=1; i<=N; i++ ) Degs.push_back(DegVar); 
      ShoupZPX ShoupZPXElm[BBSPARSEAUX_MAXTERMS];
      BBSPARSEAUX_VecDegVector PruneDegs;
      bbvector< ShoupZP >      PruneMons;

      SparseInterp( FirstFacTShoupZPMPI, Vars, Degs, DegDet, &MPCard, 
                    PruneDegs, PruneMons, ShoupZPXElm, ProbName, IsRestart,
                    NumProc );

      GetCPUTime( "toesacQXsacZPX.C" );
      ShoupCleanUpEnv();
   }
   else {
      ShoupInitEnv( &MPCard );
      BBSymToeDetShoupZP  TShoupZP( N, DegDet , Prob );
      sprintf( FileName, "%dx%dtoesacQXshoupZPX", N, N );
      BBServerShoupZPXMPI Server( BBMPI_DRIVER_RANK, MyRank, TShoupZP, 
                                  FileName, BBMPI_CLOCK_ON );
      ShoupCleanUpEnv();
   }

   mpz_clear( &MPCard );
   MPI_Finalize();
}

