
/*-------------------------------------------------------------
  module: bbnumdenaux.h
  author: Angel L. Diaz

  date: 11/1/96

  description: 
     The functions and classes in this module provide auxiliary 
     functionality utilized by the numerator/denominator separation 
     algorithm. 

  functions:
     SpecialExtGCD
     Pade 
      
  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.
 
-------------------------------------------------------------*/

#ifndef _BBNUMDENAUX
#define _BBNUMDENAUX

#define BBNUMDENAUX_MAXCOEFFS 500
#define BBNUMDENAUX_MAXVARS   500
#define BBNUMDENAUX_MAXDEG    500

/*
// for Cauchy
#define BBNUMDENAUX_MAXCOEFFS 1000
#define BBNUMDENAUX_MAXVARS   1000
#define BBNUMDENAUX_MAXDEG    1000
*/

#include "BlackBox/CommonObjects/blackbox.h"
#include "BlackBox/Aux/bbdense.h"
#include "BlackBox/Algorithms/bbnumdentrace.h"

/*-------------------------------------------------------------
  function: SpecialExtGCD
  date: 9/1/96

  reference: 
     E. Kaltofen and B. Trager, "On Computing with 
     Polynomials Given By Black Boxes for Their Evaluations: 
     Greatest Common Divisors, Factorization, Seperation of 
     Numerator and Denominators", J. Sym. Comp. (1990), 9/3,
     pg. 315. 

  description: 
     A version of the specialized extended Euclidean
     algorithm found in the above reference. Given polynomials
     H(X) and Y(X) and a degree bound d, compute polynomials 
     FHat(X) and GHat(X) such that 
           1) FHat(X) == GHat(X)*H(X) (mod Y(X))
           2) GCD( FHat(X), GHat(X) ) = 1
     and deg(FHat(X)) <= d, deg(GHat(X)) <= e, ldcf(GHat(X)) = 1

-------------------------------------------------------------*/

template < class P, class K > 
void SpecialExtGCD(
   P&           H_i,   // H(X), deg(H(X)) <= d + e
   P&           Y_i,   // Y(X) = (X - i_1) * (X - i_2) * ... (X - i_d+e+1)
   unsigned int d_i,   // d = deg( F ), e = deg( G )
   K&           lc_i,
   P&           FHat_o,
   P&           GHat_o
   )
{
   MP_INT MPOne, MPZero;
   P H0,H1,H2;
   P G0,G1,G2;
   P Q;
   mpz_init_set_ui( &MPZero, 0 );
   mpz_init_set_ui( &MPOne, 1 );

#ifdef BBDEFS_NUMDEN_TRC
   BBNUMDEMSGCDTrace1( );
#endif

   H0 = Y_i;
   H1 = H_i;
   G0 = P( (MP_INT*)&MPZero, 0 );
   G1 = P( (MP_INT*)&MPOne, 0 );

   while (  H0.dg() > d_i ) {
       H2.pquorem( H0, H1, Q, H2 );
       G2 = G0 - Q*G1;
       H0 = H1; H1 = H2;
       G0 = G1; G1 = G2;
   }

   FHat_o = (H0*lc_i)/G0.lcoeff( );
   GHat_o = (G0*lc_i)/G0.lcoeff( );

#ifdef BBDEFS_NUMDEN_TRC
   BBNUMDEMSGCDTrace2( );
#endif

   mpz_clear( &MPOne );
   mpz_clear( &MPZero );
}

/*-------------------------------------------------------------
  function: Pade
  date: 9/3/96

  reference: 
     E. Kaltofen and B. Trager, "On Computing with 
     Polynomials Given By Black Boxes for Their Evaluations: 
     Greatest Common Divisors, Factorization, Seperation of 
     Numerator and Denominators", J. Sym. Comp. (1990), 9/3,
     pg. 314--315. 

  description: 
     A version of the specialized Pade approximation
     algorithm found in the above reference. Given a black for
     f(x1,...,xn)/g(x1,...,xn), d = deg( f ), e = deg( g ) and
     n linear forms, compute the polynomials FHat(X) and 
     GHat(X).

-------------------------------------------------------------*/

template < class P, class K, class B > 
void Pade(
   B*       BB_i,        // black box for f(x1,...,xn)/g(x1,...,xn)
   unsigned int d_i,     // deg( f )
   unsigned int e_i,     // deg( g ) 
   P*       LinForms_i,  // n linear forms
   K&       LC_i,            
   K&       GenElm_m,    // distinct field element generator
   int&     StaticDeg_m, // -1 indicates refresh 
   K*       Ind_m,       // independent values X_i[1], ... ,X_i[Deg_i+1]
   K*       Reciprocal_m,
   P*       Basis_m,
   P&       FHat_o,
   P&       GHat_o
   )
{
   MP_INT MPOne, MPZero;
   mpz_init_set_ui( &MPOne, 1 );
   mpz_init_set_ui( &MPZero, 0 );
   int j,k,i;
   unsigned int NumValidPnts;
   unsigned int NumVariables;
   K Pnts[BBNUMDENAUX_MAXVARS];
   K Dep[BBNUMDENAUX_MAXCOEFFS];
   P H,Y;
   P X( (MP_INT*)&MPOne, 1 );

#ifdef BBDEFS_NUMDEN_TRC
   BBNUMDEMPADETrace1( );
#endif

   NumValidPnts = d_i + 2*e_i + 1;
   NumVariables = (*BB_i).NumVars();
   k = 0;
   for( j=0; j<NumValidPnts; j++ ) {
      if ( StaticDeg_m < j ) {
         Ind_m[j] = GenElm_m.DistinctSelect();
      }
      for( i=0; i<NumVariables; i++ ) {
         Pnts[i] = LinForms_i[i].eval( &Ind_m[j] );
      }
      // MUST ADD CHECK FOR PNT THAT MAPS g(x1,...,xn) TO ZERO
      Dep[j] = *((*BB_i)( Pnts )); 
      k++;
      if ( k == d_i + e_i + 1 )
         break;
   }

   if ( k < d_i + e_i + 1 ) {
      FHat_o = K( (MP_INT*)&MPZero );
      GHat_o = K( (MP_INT*)&MPZero );
      goto XDONE;
   }

   UniDenseInterp( Ind_m, Dep, d_i + e_i, StaticDeg_m, Reciprocal_m, 
                   Basis_m, H );  

   Y = X - Ind_m[0];
   for( i=1; i<k; i++)
      Y = Y *( X - Ind_m[i] );

   SpecialExtGCD( H, Y, d_i, LC_i, FHat_o, GHat_o );

#ifdef BBDEFS_NUMDEN_TRC
   BBNUMDEMPADETrace2( );
#endif
   XDONE:
   mpz_clear( &MPOne );
   mpz_clear( &MPZero );
} 

#endif
