/*-------------------------------------------------------------
  module: bbsparseaux.h
  author: Angel L. Diaz
  date: 8/12/96
-------------------------------------------------------------*/

#ifndef _BBSPARSEAUX
#define _BBSPARSEAUX

/*
// for Toeplitz
#define BBSPARSEAUX_MAXVARS     20
#define BBSPARSEAUX_MAXTERMS    15000
#define BBSPARSEAUX_MAXJI       15000
*/

#define BBSPARSEAUX_MAXVARS     100
#define BBSPARSEAUX_MAXTERMS    1000
#define BBSPARSEAUX_MAXJI       1000 

#define BBSPARSEAUX_RESERVEDVAR 100
#define BBSPARSEAUX_SEED        103069

/*
// for Toeplitz 
#define BBSPARSEAUX_MAXVARS     20
#define BBSPARSEAUX_MAXTERMS    15000
#define BBSPARSEAUX_MAXJI       15000 
*/

typedef bbsvector< unsigned int > BBSPARSEAUX_DegVector;
typedef bbsvector< unsigned int > BBSPARSEAUX_VarVector;
typedef bbvector< BBSPARSEAUX_DegVector > BBSPARSEAUX_VecDegVector;

#include "BlackBox/Aux/bbdefs.h"
#include "BlackBox/Aux/bbtime.h"
#include "BlackBox/Algorithms/bbsparsetrace.h"

/*-------------------------------------------------------------
  function: VandermondeSolveZippel
  date: 8/12/96

  reference: R. Zippel, "Interpolating polynomials from their values", 
     J. Symbolic Comput., 9/3, 1990 pg. 375--403

  description: Richard Zippel's algorithm for solving a transposed 
     vandermonde system of equations ( A x = a ). Variable names mirror 
     thoses utilized in Kaltofen's and Lakshman's paper referenced below.

  return: Vector x = transpose(x1,...,xn_i), where x = inverse(A)*a
-------------------------------------------------------------*/

template< class K, class P > inline void VandermondeSolveZippel(
   bbvector< K >& v_i,   // vector which describes a Vandermonde matrix A 
   bbvector< K >& a_i,   // vector a = transpose(a1,...,an_i) 
   bbvector< K >& x_o,   // solution vector 
   P* B_o
   )
{
   unsigned int j,k,i;
   unsigned int n; 
   MP_INT MPOne, MPZero;
   mpz_init_set_ui( &MPOne, 1 );
   mpz_init_set_ui( &MPZero, 0 );
   P Z( &MPOne, 1);
   K tmp;
   n = v_i.size();

   // Compute the polynomials Bj(z)
   for( j=0; j<n ;j++ ) {
      B_o[j] = (P)&MPOne;
      for( k=0; k<n; k++ ) {
         if ( j != k ) {
            B_o[j] = B_o[j] * ( ( Z - v_i[k] ) / ( v_i[j] - v_i[k] ) );
         } 
      }
   }

   while ( x_o.size() > 0 ) x_o.pop_back();
   // Compute the solution vector from the coefficients of Bj(z)
   for( j=0; j<n; j++ ) {
      tmp = (K)&MPZero;
      for( i=0; i<n; i++ ) 
         tmp = tmp + (B_o[j])[i] * a_i[i];
      x_o.push_back( tmp );
   }

   mpz_clear( &MPOne );
   mpz_clear( &MPZero );
}

/*-------------------------------------------------------------
  function: VandermondeSolveKL
  date: 8/12/96

  reference: E. Kaltofen and Yagati Lakshman, "Improved sparse 
     multivariate polynomial interpolation algorithms",Proc. ISSAC '88, 
     Springer Lect. Notes Comput. Sci., P. Gianni ed., pg. 467--474 

  description: Kaltofen's and Lakshman's efficient algorithm to solve 
     a transposed vandermonde system of equations ( A x = a ). Variable
     names mirror thoses utilized in Kaltofen's and  Lakshman's paper.
 
  return: Vector x = transpose(x1,...,xn_i), where x = inverse(A)*a
-------------------------------------------------------------*/

template< class K, class P > inline void VandermondeSolveKL(
   bbvector< K >& v_i,   // vector which describes a Vandermonde matrix A 
   bbvector< K >& a_i,   // vector a = transpose(a1,...,an_i) 
   bbvector< K >& x_o,   // solution vector 
   P* B_o
   )
{
   MP_INT MPOne;
   mpz_init_set_ui( &MPOne, 1 );
   unsigned int i;
   unsigned int n; 
   P B,BP;
   P D;
   P BD;
   P QN;
   n = v_i.size();

   // compute the polynomials B(z) and BP(z)
   B = P( &MPOne, 1) - v_i[0];
   for( i=1; i<n; i++ ) 
      B = B * ( P( &MPOne, 1) - v_i[i] );
   BP = B.Diff();

   // compute the polynomials D(z) and B(z)*D(z)
   D = P( &MPOne, n ) * a_i[0]; 
   for( i=1; i<n; i++ ) 
      D = D + P( &MPOne, n - i ) * a_i[i];
   BD = B * D;

   // compute the polynomial QN(w)
   QN = BD[ n + 1 ];
   for( i=n+2; i<=2*n; i++ )  
      QN = QN + P( &MPOne, i - n - 1 ) * BD[i];

   while ( x_o.size() > 0 ) x_o.pop_back();
   // compute solution vector
   for( i=0; i<n; i++ ) 
      x_o.push_back( QN.eval( &v_i[i] ) / BP.eval( &v_i[i] ) ); 

   mpz_clear( &MPOne );
}

/*-------------------------------------------------------------
  function: Max

  description: 

  returns: number of variables in augmented variable array
-------------------------------------------------------------*/

inline unsigned int MaxDegree(
   BBSPARSEAUX_DegVector& DegArray_i
   )
{
   unsigned int i;
   unsigned int Max;
  
   Max = DegArray_i[0];
   for( i=0; i<DegArray_i.size() ;i++ )
      if ( DegArray_i[i] > Max ) 
         Max = DegArray_i[i];
   
   return Max;
} 
/*-------------------------------------------------------------
  function: ProdRaisedPnts
  date: 8/13/96

  reference: this corresponds to the product b_k,1^e1*...b_k,i^d described
     on page 389 of "Factorization of polynomials given by straight-line 
     programs", E. Kaltofen, Randomness and Computation Vol 5, Advances 
     in Computing Research, S. Micali ed., Greenwhich, Connecticut, 1989,
  
  description: compute the product of each of the points raised to its
     corresponding exponent. The last point (representing the next
     varible to interpolate) is raised to d_i.

  returns: pnts_i[1]^(exps[1]) * ... * pnts_i[i_i-1]^(exps[i_i-1])* 
           pnts_i[i_i]^(d_i)
-------------------------------------------------------------*/

template< class K > inline K ProdRaisedPnts(   
   K*                     pnts_i, // array of points 
   BBSPARSEAUX_DegVector& exps_i, // vector of corresponding exponents
   unsigned int           d_i,    // exponent of last point 
   unsigned int           i_i     // number of pnts to "raise" 
   )
{
   unsigned int j;
   K Prod;
 
   if ( i_i == 1 )
      Prod = pnts_i[0].Exp( d_i );
   else {           
      Prod = pnts_i[0].Exp( exps_i[0] ); 
      for( j=1; j < i_i - 1; j++ )
         Prod = Prod * pnts_i[j].Exp( exps_i[j] );
      Prod = Prod * pnts_i[i_i-1].Exp( d_i );
   }
  
   return Prod;
}

/*-------------------------------------------------------------
  function: RefineJi
  date: 8/13/96

  description: Select exponents from the full set of exponent vectors
     which have a corresponding non zero solution in the solution 
     vector
-------------------------------------------------------------*/

template< class K > inline void RefineJi(
   bbvector< K >&             x_i,    // solution vector 
   unsigned int               n_i,    // length of solution vector 
   BBSPARSEAUX_VecDegVector&  nJi_i,  // full vector of exponent vectors 
   BBSPARSEAUX_VecDegVector&  Ji_o    // refined vector of exponent vectors 
   )
{
   unsigned int i;
   MP_INT MPZero;
   mpz_init_set_ui( &MPZero, 0 );

   while ( Ji_o.size() > 0 ) Ji_o.pop_back();
   for( i=0 ;i < n_i ;i++ ) {
      if ( x_i[i] != (MP_INT *) &MPZero ) {
         Ji_o.push_back( nJi_i[i] );
      }
   } 

   mpz_clear( &MPZero );
}

/*-------------------------------------------------------------
  function:CompPruneDeg
  date: 8/13/96

  description: compute the total degree of the remaining variables
     in a monomial given an exponent vector. The fist entry in 
     the exponent vector is the total degree of the monomial and 
     the remaining entries are the degrees of the variables 
     interpolated thus far.
 
  returns: a non negative integer, total degree of the remaining 
     variables
-------------------------------------------------------------*/

inline unsigned int CompPruneDeg(
   BBSPARSEAUX_DegVector& DegVector_i
   )
{
   unsigned int i,d;

   // the first entry is the total degree of the monomial (extra variable)
   d = DegVector_i[0];

   // compute the total degree of the remaining variables (if 0 we can prune)
   for(i=1;i<DegVector_i.size();i++) {
      d = d - DegVector_i[ i ];
   }

   return( d );
}

/*-------------------------------------------------------------
  function: HomoEvalBB
  date: 8/14/96

  description: evaluate a black box utilizing PntArray_i[0]
     as the homogenizing variable. 
-------------------------------------------------------------*/

template < class K, class B > inline K HomoEvalBB(
   B& bb_i,
   K* PntArray_i,
   unsigned int NumPnts_i
   )
{
   unsigned int i;
   K NewPntArray[BBSPARSEAUX_MAXVARS];    // adjusted list of points
   
   for( i=1; i<NumPnts_i; i++ ) {
      NewPntArray[i-1] = PntArray_i[0] * PntArray_i[i];
   }

   return ( *(bb_i( NewPntArray )) );
}

/*-------------------------------------------------------------
  function: HomoDistBB
  date: 9/21/96

  description: evaluate a black box utilizing PntArray_i[0]
     as the homogenizing variable. 
-------------------------------------------------------------*/

template < class K, class B > inline void HomoDistBB(
   int          Dest_i,
   unsigned int NumPnts_i,
   B&           BB_i,
   K*           PntArray_i,
   int&         IsDist_m,
   int&         Status_m
   )
{
   unsigned int i;
   K NewPntArray[BBSPARSEAUX_MAXVARS];    // adjusted list of points
   
   for( i=1; i<NumPnts_i; i++ ) {
      NewPntArray[i-1] = PntArray_i[0] * PntArray_i[i];
   }

   BB_i.Distribute( Dest_i, NewPntArray, IsDist_m, Status_m );
}

/*-------------------------------------------------------------
  function: HomoEvalPoly
  date: 8/19/96

  description: evaluate a polynomial given by a vector of monomials
     and a vector of exponents utilizing PntArray_i[0] as the 
     homogenizing variable. 
-------------------------------------------------------------*/

template < class K > inline K HomoEvalPoly(
   K*                        PntArray_i, // evaluation points
   unsigned int              NumPnts_i,
   BBSPARSEAUX_VecDegVector& PruneDegs_i,// exponent vector of pruned poly
   bbvector< K >&            PruneMons_i // monomial vector of pruned poly
   )
{
   MP_INT MPZero;
   mpz_init_set_ui( &MPZero, 0 );
   K Value = (K)&MPZero;
   K C;
   unsigned int i, j;
   unsigned int DegsSize, MonsSize;
 
   MonsSize = PruneMons_i.size();

   // no pruned polynoials
   if ( MonsSize == 0 ) {
      Value = K( &MPZero );
      goto XDONE;  
   }

   // loop through monomials
   for( i=0; i<MonsSize; i++ ) {
      C = PntArray_i[0].Exp( (PruneDegs_i[i])[0] );
      DegsSize = (PruneDegs_i[i]).size();
      // loop through exponents
      for( j=1; j<DegsSize; j++ ) 
         C = C * PntArray_i[j].Exp( (PruneDegs_i[i])[j] );
       
      Value = Value + C * PruneMons_i[i];
   }
   
   XDONE:
   mpz_clear( &MPZero );
   return Value;
}

/*-------------------------------------------------------------
  function: Prune
  date: 8/14/96

  description: 
-------------------------------------------------------------*/

template< class K > inline void Prune(
   BBSPARSEAUX_VarVector&    Vars_i, // vector of variable numbers
   bbvector< unsigned int >& ell_m,  // "remaining degrees" for each monomial
   BBSPARSEAUX_VecDegVector& Ji_m,   // vector of exponent vectors
   bbvector< K >&            x_m,    // solution vector 
   BBSPARSEAUX_VecDegVector& PruneDegs_m, // exponent vector of pruned poly
   bbvector< K >&            PruneMons_m  // monomial vector of pruned poly
   )
{
   unsigned int i;
   MP_INT MPZero;
   mpz_init_set_ui( &MPZero, 0 );
   bbvector< K > MonomialVector;
   bbvector< unsigned int > NewEll; 
   BBSPARSEAUX_VecDegVector NewJi;

   // refine the solution vector
   for( i = 0; i < x_m.size(); i++ )
      if ( x_m[i] != (MP_INT*) &MPZero ) 
         MonomialVector.push_back( x_m[i] );

    while ( x_m.size() > 0 ) x_m.pop_back();
    x_m.insert( x_m.begin(), MonomialVector.begin(), MonomialVector.end() );

    // update monomial degree and exponent vectors
    for( i=0; i < ell_m.size(); i++ ) {
       // if this monomial can be pruned
       if ( ell_m[i] == 0 ) {
          PruneMons_m.push_back( MonomialVector[i] ); 
          PruneDegs_m.push_back( Ji_m[i] );
       }
       else {
          NewEll.push_back( ell_m[i] );
          if ( Ji_m.size() > 0 )
             NewJi.push_back( Ji_m[i] ); 
       } 
    }

    while ( ell_m.size() > 0 ) ell_m.pop_back();
    ell_m.insert( ell_m.begin(), NewEll.begin(), NewEll.end() );

    while ( Ji_m.size() > 0 ) Ji_m.pop_back();
    Ji_m.insert( Ji_m.begin(), NewJi.begin(), NewJi.end() );
  
    mpz_clear( &MPZero );

}

/*-------------------------------------------------------------
  function: LoadState
  date: 12/5/96

  description: 
-------------------------------------------------------------*/

template< class K >
void LoadState(
   unsigned int&              BBCnt_m,
   unsigned int&              T_m,
   K*                         InputPnts_m,
   unsigned int               NumVars_i,
   BBSPARSEAUX_VecDegVector&  Ji_m,
   bbvector< K >&             X_m,
   BBSPARSEAUX_VecDegVector&  PruneDegs_m, 
   bbvector< K >&             PruneMons_m,
   char*                      ProbName_i, 
   int&                       VarID_m
   )
{
   int i,j,d,d1,d2;
   char FileName[BBDEFS_FILEZ];  
   char KElm[BBDEFS_KSTRZ];
   BBSPARSEAUX_DegVector tmp;

   sprintf( FileName, "%s%d.chk", ProbName_i, VarID_m );
   ifstream InFile( FileName, ios::in );
   if ( !InFile ) {
      cerr << "LoadState : cannot open " << FileName
           << " for input" << endl;
      exit(-1);
   }
   InFile >> BBCnt_m;
   InFile >> T_m;
   InFile >> NumVars_i;
   for(i=0; i<NumVars_i; i++ ) {
      InFile >> KElm;
      InputPnts_m[i] = K( KElm );
   }
   // Ji
   InFile >> d1;
   for( i=0; i<d1; i++ ) {
      InFile >> d2;
      while ( tmp.size() > 0 ) tmp.pop_back();
      for( j=0; j<d2; j++ ) {
         InFile >> d;
         tmp.push_back( d );
      }
      Ji_m.push_back( tmp );
   }
   // X
   InFile >> d1;
   for( i=0; i<d1; i++ ) {
      InFile >> KElm;
      X_m.push_back( K( KElm ) ); 
   }
   // PruneDegs
   InFile >> d1;
   for( i=0; i<d1; i++ ) {
      InFile >> d2;
      while ( tmp.size() > 0 ) tmp.pop_back();
      for( j=0; j<d2; j++ ) {
         InFile >> d;
         tmp.push_back( d );
      }
      PruneDegs_m.push_back( tmp );
   }
   // PruneMons 
   InFile >> d1;
   for( i=0; i<d1; i++ ) {
      InFile >> KElm;
      PruneMons_m.push_back( K( KElm ) );
   }

   InFile.close();

   VarID_m = BBDEFS_FALSE;
}

/*-------------------------------------------------------------
  function: SaveState
  date: 12/5/96

  description:
-------------------------------------------------------------*/
template< class K >
void SaveState(
   unsigned int              BBCnt_i,
   unsigned int              T_i,
   K*                        InputPnts_i,
   unsigned int              NumVars_i,
   BBSPARSEAUX_VecDegVector& Ji_i,
   bbvector< K >&            X_i,
   BBSPARSEAUX_VecDegVector& PruneDegs_i, 
   bbvector< K >&            PruneMons_i,
   char*                     ProbName_i, 
   int                       VarID_i
   )
{
   int i,j;
   char FileName[BBDEFS_FILEZ];  
   char KElm[BBDEFS_KSTRZ];

   sprintf( FileName, "%s%d.chk", ProbName_i, VarID_i );
   ofstream OutFile( FileName, ios::out );
   if ( !OutFile ) {
      cerr << "SaveState : cannot open " << FileName
           << " for output" << endl;
      exit(-1);
   }

   OutFile << BBCnt_i      << endl;
   OutFile << T_i          << endl;
   OutFile << NumVars_i    << endl;
   for(i=0; i<NumVars_i; i++ ) {
      InputPnts_i[i].toString( KElm ); 
      OutFile << KElm << endl;
   }
   // Ji
   OutFile << Ji_i.size()  << endl; 
   for( i=0; i<Ji_i.size(); i++ ) {
      OutFile << Ji_i[i].size() << endl;
      for( j=0; j<Ji_i[i].size(); j++ ) {
         OutFile << Ji_i[i][j] << endl;
      }
   }
   // X
   OutFile << X_i.size()  << endl; 
   for( i=0; i<X_i.size(); i++ ) {
      X_i[i].toString( KElm ); 
      OutFile << KElm << endl;
   }
   // PruneDegs
   OutFile << PruneDegs_i.size()  << endl; 
   for( i=0; i<PruneDegs_i.size(); i++ ) {
      OutFile << PruneDegs_i[i].size() << endl;
      for( j=0; j<PruneDegs_i[i].size(); j++ ) {
         OutFile << PruneDegs_i[i][j] << endl;
      }
   }
   // PruneMons 
   OutFile << PruneMons_i.size()  << endl; 
   for( i=0; i<PruneMons_i.size(); i++ ) {
      PruneMons_i[i].toString( KElm ); 
      OutFile << KElm << endl;
   }

   OutFile.close();
}

#endif
