
/*-------------------------------------------------------------
  module: bbpoly.h
  author: Angel L. Diaz

  date: 10/21/96

  description: 
     The classes in this module realize the black box 
     representation of common object polynomials. In the FoxBox 
     library, a polynomial black box object is a function object which 
     when supplied with values (a pointer to an array of field elements) 
     for the symbols, computes the value for the polynomial. 

  classes: 
     BlackBoxPoly 
     BlackBoxUPoly
     BlackBoxUPolyEx

  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.

-------------------------------------------------------------*/

#ifndef _BBPOLY
#define _BBPOLY

#include "BlackBox/CommonObjects/blackbox.h"

/*-------------------------------------------------------------
  class: BlackBoxPoly 
  date: 10/21/96

  description: 
     Provides for constructing a common object black box 
     polynomial from a native imported domain arithmetic
     polynomial (K and KPoly).

-------------------------------------------------------------*/

template < class K, class KPoly > 
class BlackBoxPoly : public BlackBox< K > {
   private:
      KPoly Polynomial;
      K     Value;
   public:
      BlackBoxPoly( );
      BlackBoxPoly( const BlackBoxPoly< K, KPoly >& );
      BlackBoxPoly( const KPoly&, unsigned int, unsigned int, double );
      ~BlackBoxPoly( ) { }

      inline K* operator () ( K* );
      inline BlackBoxPoly& operator=( const BlackBoxPoly< K, KPoly >& );
};

template< class K, class KPoly > 
BlackBoxPoly< K, KPoly >::BlackBoxPoly( ) : BlackBox< K >( ) {
}

template< class K, class KPoly > 
BlackBoxPoly< K, KPoly >::BlackBoxPoly( 
   const BlackBoxPoly< K, KPoly >& Poly_i
   ) : BlackBox< K >( )
{
   *this = Poly_i;
}

template< class K, class KPoly > 
BlackBoxPoly< K, KPoly >::BlackBoxPoly(
   const KPoly&  Poly_i,
   unsigned int  NumVars_i,
   unsigned int  Deg_i,
   double        Prob_i
   ) : BlackBox< K >( )
{
   Polynomial   = Poly_i;
   NumVariables = NumVars_i;
   Degree       = Deg_i;
   Probability  = Prob_i;
}

template < class K,  class KPoly > 
inline K* BlackBoxPoly< K, KPoly >::operator () (
   K* Pnts_i
   )
{
   Value = Polynomial.eval( Pnts_i );
   return &Value;
}

template< class K, class KPoly > 
inline BlackBoxPoly< K, KPoly >& BlackBoxPoly< K, KPoly>::operator=(
   const BlackBoxPoly< K, KPoly >& Poly_i
   )
{
   Polynomial   = Poly_i.Polynomial;
   NumVariables = Poly_i.NumVariables;
   Degree       = Poly_i.Degree;
   Probability  = Poly_i.Probability;

   return *this;
}

/*-------------------------------------------------------------
  class: BlackBoxUPoly 
  date: 10/21/96

  description:
     Provides for constructing a common object black box 
     polynomial from a pointer to a function. It is expected 
     that the "unknown" function takes as input an array of
     coefficient domain elements (K) and returns a coefficient 
     domain element.

-------------------------------------------------------------*/

template< class K > 
class BlackBoxUPoly : public BlackBox< K > {
   private:
      typedef K ( *FPTRK )( K* );
      FPTRK Polynomial;
      K Value;
   public:
      BlackBoxUPoly( );
      BlackBoxUPoly( const BlackBoxUPoly< K >& );
      BlackBoxUPoly( FPTRK         PolyK_i,
                     unsigned int NumVars_i,
                     unsigned int Deg_i,
                     double       Prob_i
                   ) : BlackBox< K >()
      {
         Polynomial   = PolyK_i;
         NumVariables = NumVars_i;
         Degree       = Deg_i;
         Probability  = Prob_i;
      }

      ~BlackBoxUPoly() { }

      inline K* operator () ( K* );
      inline BlackBoxUPoly< K >& operator=( const BlackBoxUPoly< K >& );
};

template< class K > 
BlackBoxUPoly< K >::BlackBoxUPoly( ) : BlackBox< K >( ) {
}

template< class K > 
BlackBoxUPoly< K >::BlackBoxUPoly(
   const BlackBoxUPoly<K>& Poly_i
   ) : BlackBox< K >()
{
   *this = Poly_i;
}

template< class K > 
inline K* BlackBoxUPoly< K >::operator () (
   K *Pnts_i
   )
{
   Value = ( *Polynomial )( Pnts_i );
   return &Value;
}

template< class K > 
inline BlackBoxUPoly< K >& BlackBoxUPoly< K >::operator=(
   const BlackBoxUPoly< K >& Poly_i
   )
{
   Polynomial   = Poly_i.Polynomial;
   NumVariables = Poly_i.NumVariables;
   Degree       = Poly_i.Degree;
   Probability  = Poly_i.Probability;

   return *this;
}

/*-------------------------------------------------------------
  class: BlackBoxUPolyEx
  date: 10/21/96

  description: 
     Provides for constructing a common object black box 
     polynomial from a pointer to a function. It is expected 
     that two "unknown" functions are provided:

      (1) takes as input an array of coefficient domain 
          elements (K) and returns a coefficient domain element.

      (2) takes as input an array of coefficient extension 
          domain elements (L) and returns a coefficient 
          extension domain element.

-------------------------------------------------------------*/

template< class K, class L > 
class BlackBoxUPolyEx : public BlackBoxEx< K, L > {
   private:
      typedef K ( *FPTRK )( K* );
      typedef L ( *FPTRL )( L* ); 
      FPTRK PolynomialK;
      FPTRL PolynomialL;
      K KValue;
      L LValue;
   public:
      BlackBoxUPolyEx( );
      BlackBoxUPolyEx( const BlackBoxUPolyEx< K, L >& );
      BlackBoxUPolyEx( FPTRK         PolyK_i, 
                       FPTRL         PolyL_i, 
                       unsigned int  NumVars_i, 
                       unsigned int  Deg_i,
                       double        Prob_i
                   ) : BlackBoxEx< K, L >( ) 
      { 
         PolynomialK  = PolyK_i; 
         PolynomialL  = PolyL_i;
         NumVariables = NumVars_i; 
         Degree       = Deg_i; 
         Probability  = Prob_i;
      }

      ~BlackBoxUPolyEx() { }

      inline K* operator () ( K* );
      inline L* operator () ( L* );
      inline BlackBoxUPolyEx<K,L>& operator=( const BlackBoxUPolyEx<K,L>& );
};

template< class K, class L > 
BlackBoxUPolyEx< K, L >::BlackBoxUPolyEx( ) : BlackBoxEx< K, L >() {
}

template< class K, class L > 
BlackBoxUPolyEx< K, L >::BlackBoxUPolyEx( 
   const BlackBoxUPolyEx< K, L >& Poly_i 
   ) : BlackBoxEx< K, L >() 
{ 
   *this = Poly_i; 
}

template< class K, class L > 
inline K* BlackBoxUPolyEx< K, L >::operator () (
   K *Pnts_i
   ) 
{ 
   KValue = ( *PolynomialK )( Pnts_i ); 
   return &KValue;
}

template< class K, class L > 
inline L* BlackBoxUPolyEx< K, L >::operator () (
   L *Pnts_i
   )
{
   LValue = ( *PolynomialL )( Pnts_i );
   return &LValue;
}

template< class K, class L > 
inline BlackBoxUPolyEx< K, L >& BlackBoxUPolyEx< K, L >::operator=(
   const BlackBoxUPolyEx< K, L >& Poly_i
   )
{ 
   PolynomialK  = Poly_i.PolynomialK; 
   PolynomialL  = Poly_i.PolynomialL;
   NumVariables = Poly_i.NumVariables;
   Degree       = Poly_i.Degree; 
   Probability  = Poly_i.Probability;

   return *this; 
}

#endif
