
/*-------------------------------------------------------------
  module: bbfachmap.h
  author: Angel L. Diaz

  date: 10/24/96

  description: 
     The class in this module realizes the homorphic map of a black 
     box factorization black box. 

  class:
     BlackBoxFactorsHMap

  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.
 
-------------------------------------------------------------*/

#ifndef __BBFACHMOD
#define __BBFACHMOD

#include "BlackBox/HomomorphicMaps/bbhmap.h"
#include "BlackBox/HomomorphicMaps/bbfachmaptrace.h"
#include "BlackBox/Algorithms/bbfactortrace.h"
#include "BlackBox/Algorithms/bbfactoraux.h"

/*-------------------------------------------------------------
  class: BlackBoxFactorsHMap
  date: 10/24/96

  description:
     Provides for constructing a homorphic map of a factors 
     black box object.

-------------------------------------------------------------*/

template < class K, class KP, class F, class FP, class B, class FB, class H > 
class BlackBoxFactorsHMap : public BlackBox< F > {
   protected:
      H h;
      FB* BB;
      BlackBoxFactors<K,KP,B>* BBF;
      int ConvertStatus;
      F RandA[BBFACTOR_MAXVAR];
      F RandB[BBFACTOR_MAXVAR];
      F RandC[BBFACTOR_MAXVAR];
      F ProbeElm;
      F GenElm;
      int StaticDeg;
      F Ind[BBDENSE_MAXCOEFFS];
      F Reciprocal[BBDENSE_MAXCOEFFS];
      FP Basis[BBDENSE_MAXCOEFFS];
      unsigned int NumFactors;
      unsigned int SumExps;
      unsigned int Exps[BBFACTOR_MAXFACTORS];
      int Sign;
      unsigned int DegF;
      int ParFracR;
      int ParFracC;
      F Cont;
      FP FactorImages[BBFACTOR_MAXFACTORS];
      FP FactorProducts[BBFACTOR_MAXFACTORS*BBFACTOR_MAXDEG];
      FP pgsem1;
      FP ParFrac[BBFACTOR_MAXDEG][BBFACTOR_MAXFACTORS];
      F ValueOfFactors[BBFACTOR_MAXFACTORS];
   public:
      inline friend ofstream& operator<<(ofstream&, 
         BlackBoxFactorsHMap<K,KP,F,FP,B,FB,H>& );
      inline friend ifstream& operator>>(ifstream&, 
         BlackBoxFactorsHMap<K,KP,F,FP,B,FB,H>& );
     
      BlackBoxFactorsHMap( ) { }
      BlackBoxFactorsHMap( FB& );
      BlackBoxFactorsHMap( FB&, BlackBoxFactors<K,KP,B>&, H&);
      void Construct( FB&, BlackBoxFactors<K,KP,B>&, H& );
      ~BlackBoxFactorsHMap( ); 

      inline F* operator () ( F* );
      inline int IsConverted() { return ConvertStatus; }

      // static information access functions
      inline int GetStaticDeg()            { return StaticDeg;         }
      inline int GetNumFactors()           { return NumFactors;        }
      inline int GetSumExps()              { return SumExps;           }
      inline int GetSign()                 { return Sign;              }
      inline int GetDegF()                 { return DegF;              }
      inline int GetParFracR()             { return ParFracR;          }
      inline int GetParFracC()             { return ParFracC;          }
      inline int GetExps( int i )          { return Exps[i];           }
      inline F  GetRandA( int i )          { return RandA[i];          }
      inline F  GetRandB( int i )          { return RandB[i];          }
      inline F  GetRandC( int i )          { return RandC[i];          }
      inline F  GetProbeElm()              { return ProbeElm;          }
      inline F  GetGenElm()                { return GenElm;            }
      inline F  GetInd( int i )            { return Ind[i];            }
      inline F  GetReciprocal( int i )     { return Reciprocal[i];     }
      inline FP GetBasis( int i )          { return Basis[i];          }
      inline F  GetCont()                  { return Cont;              }
      inline FP GetFactorImages( int i )   { return FactorImages[i];   }
      inline FP GetFactorProducts( int i)  { return FactorProducts[i]; }
      inline FP Getpgsem1()                { return pgsem1;            }
      inline FP GetParFrac( int i, int j ) { return ParFrac[i][j];     }
};

template < class K, class KP, class F, class FP, class B, class FB, class H >
BlackBoxFactorsHMap<K,KP,F,FP,B,FB,H>::BlackBoxFactorsHMap( 
   FB& FB_i
   ) 
{
   BB            = &FB_i;
}

template < class K, class KP, class F, class FP, class B, class FB, class H >
BlackBoxFactorsHMap<K,KP,F,FP,B,FB,H>::BlackBoxFactorsHMap(
   FB& FB_i,
   BlackBoxFactors<K,KP,B>& BB_i,
   H& h_i
   )
{
   Construct( FB_i, BB_i, h_i );
}

template < class K, class KP, class F, class FP, class B, class FB, class H >
void BlackBoxFactorsHMap<K,KP,F,FP,B,FB,H>::Construct(
   FB& FB_i,
   BlackBoxFactors<K,KP,B>& BB_i,
   H& h_i
   )
{
   int i,j;
   MP_INT MPZero, MPOne;
   K  KElm;
   F  FElm;
   KP KPoly;
   FP FPoly;
   HPoly< K, KP, F, FP, H > hp( h );

   mpz_init_set_ui( &MPZero, 0 );
   mpz_init_set_ui( &MPOne, 1 );
   ConvertStatus = BBDEFS_TRUE;
   BB            = &FB_i;
   BBF           = &BB_i;
   h             = h_i;
   Degree        = BB_i.Deg();
   NumVariables  = BB_i.NumVars();
   Probability   = BB_i.Prob();
   NumFactors    = BB_i.GetNumFactors(); 
   SumExps       = BB_i.GetSumExps();
   Sign          = BB_i.GetSign();
   DegF          = BB_i.GetDegF();  
   ParFracR      = BB_i.GetParFracR();
   ParFracC      = BB_i.GetParFracC();

   for( i=0; i<NumFactors; i++ ) {
      Exps[i] = BB_i.GetExps( i );
   }

   KElm          = BB_i.GetProbeElm();
   FElm          = h( KElm );
   if ( h.IsConverted() == BBDEFS_FALSE ) {
      ConvertStatus = BBDEFS_FALSE;
      goto XDONE;
   }
   ProbeElm      = FElm; 

   // reset static interpolation information
   StaticDeg     = -1;
   GenElm        = F( (MP_INT*) &MPOne );

#ifdef BBDEFS_BBFACTOR_HMAP_TRC
   BBFACSHMTrace1( Degree, NumVariables, Probability, NumFactors, SumExps, 
                   Sign, DegF, ParFracR, ParFracC, Exps, StaticDeg, ProbeElm, 
                   GenElm );
#endif

   // convert random points ai from a0,...,an-1
   // convert random points ci from c0,....,cn-1
   for( i=0; i<NumVariables-1; i++ ) {
      KElm = BB_i.GetRandA( i );
      FElm = h( KElm );
      if ( h.IsConverted() == BBDEFS_FALSE ) {
         ConvertStatus = BBDEFS_FALSE;
         goto XDONE; 
      }
      else
        RandA[i] = FElm;
      KElm = BB_i.GetRandC( i );
      FElm = h( KElm );
      if ( h.IsConverted() == BBDEFS_FALSE ) {
         ConvertStatus = BBDEFS_FALSE;
         goto XDONE;
      }
      else
        RandC[i] = FElm;
   }

   // convert random points bi from b0,...,bn
   for( i=0; i<NumVariables; i++ ) {
      KElm = BB_i.GetRandB( i );
      FElm = h( KElm );
      if ( h.IsConverted() == BBDEFS_FALSE ) {
         ConvertStatus = BBDEFS_FALSE;
         goto XDONE;
      }
      else
        RandB[i] = FElm;
   }

#ifdef BBDEFS_BBFACTOR_HMAP_TRC
   BBFACSHMTrace2( RandA, RandB, RandC, NumVariables );
#endif

   KElm          = BB_i.GetCont();
   FElm          = h( KElm );
   if ( h.IsConverted() == BBDEFS_FALSE ) {
      ConvertStatus = BBDEFS_FALSE;
      goto XDONE;
   }
   Cont          = FElm;
   KPoly          = BB_i.Getpgsem1();
   FPoly          = hp( KPoly );

   if ( hp.IsConverted() == BBDEFS_FALSE ) {
      ConvertStatus = BBDEFS_FALSE;
      goto XDONE;
   }
   pgsem1        = FPoly;

#ifdef BBDEFS_BBFACTOR_HMAP_TRC
   BBFACSHMTrace3( Cont, pgsem1 );
#endif

   // convert factor images
   for( i=0; i<NumFactors; i++ ) {
      KPoly = BB_i.GetFactorImages( i );
      FPoly = hp( KPoly );
      if ( hp.IsConverted() == BBDEFS_FALSE ) {
          ConvertStatus = BBDEFS_FALSE;
          goto XDONE;
      }
      FactorImages[i] = FPoly;
   }

   // convert factor products
   for( i=0; i<SumExps; i++ ) {
      KPoly = BB_i.GetFactorProducts( i );
      FPoly = hp( KPoly );
      if ( hp.IsConverted() == BBDEFS_FALSE ) {
          ConvertStatus = BBDEFS_FALSE;
          goto XDONE;
      }
      FactorProducts[i] = FPoly;
   }

#ifdef BBDEFS_BBFACTOR_HMAP_TRC
   BBFACSHMTrace4( FactorImages, NumFactors, FactorProducts, SumExps );
#endif

  // convert partial fractions
  for( i=0; i<=ParFracC; i++ ) {
      for( j=0; j<=ParFracR; j++ ) {
         KPoly = BB_i.GetParFrac( i, j );
         FPoly = hp( KPoly );
         if ( hp.IsConverted() == BBDEFS_FALSE ) {
            ConvertStatus = BBDEFS_FALSE;
            goto XDONE;
         }
         ParFrac[i][j] = FPoly;
      }
  }

#ifdef BBDEFS_BBFACTOR_HMAP_TRC
   BBFACSHMTrace5( ParFrac, ParFracR, ParFracC );
#endif

   XDONE:
   mpz_clear( &MPZero );
   mpz_clear( &MPOne );
}

template < class K, class KP, class F, class FP, class B, class FB, class H >
BlackBoxFactorsHMap<K,KP,F,FP,B,FB,H>::~BlackBoxFactorsHMap( )
{
}

template < class K, class KP, class F, class FP, class B, class FB, class H >
inline F* BlackBoxFactorsHMap<K,KP,F,FP,B,FB,H>::operator() (
   F* Pnts_i
   )
{
   int          i,j;
   BiPoly< FP > FBar;
   BiPoly< FP > BiFactors[BBFACTOR_MAXFACTORS];
   F            EvalPnt;

#ifdef BBDEFS_BBFACTOR_HMAP_TRC
   BBFACSTrace6( Pnts_i, NumVariables );
#endif

   MapFtoFBar<F,FP,FB> BBFBar( BB, RandA, RandB, Pnts_i, NumVariables );

/*
   // general purpose bi-variate interpolation (w/degree guessing)
   BiDenseInterp( &BBFBar, ProbeElm, GenElm, StaticDeg, Ind,
                  Reciprocal, Basis, FBar );
*/

   // special pruning bi-variate interpolation
   SpecialBiDenseInterp( &BBFBar, DegF, GenElm, StaticDeg, Ind, Reciprocal,
                         Basis, FBar );

#ifdef BBDEFS_BBFACTOR_HMAP_TRC
   BBFACSTrace7( FBar, FBar.Deg() );
#endif

   Lift( FBar, FactorImages, NumFactors, FactorProducts, SumExps, Exps,
         ParFrac, pgsem1, BiFactors );

#ifdef BBDEFS_BBFACTOR_HMAP_TRC
   BBFACSTrace8( BiFactors, NumFactors );
#endif

   EvalPnt = Pnts_i[0] - RandB[0];

   for( i=0; i<NumFactors; i++) {
      ValueOfFactors[i] = (BiFactors[i])[0].eval( &EvalPnt );
      for( j=1; j<(BiFactors[i]).Deg(); j++ )
          ValueOfFactors[i] = ValueOfFactors[i] +
              (BiFactors[i])[j].eval( &EvalPnt );
      //ValueOfFactors[i] = ValueOfFactors[i] * (long)Sign;
   }

#ifdef BBDEFS_BBFACTOR_HMAP_TRC
   BBFACSTrace9( ValueOfFactors, NumFactors );
#endif

   return( ValueOfFactors );
}

template < class K, class KP, class F, class FP, class B, class FB, class H > 
inline ofstream& operator<<(
   ofstream& of_i,
   BlackBoxFactorsHMap<K,KP,F,FP,B,FB,H>& BB_i
   )
{
   int  i,j,k;
   char ZPElm[MOD_KSTRZ];

   of_i << BBDEFS_FACTOR_HMAP_ID << endl;
   of_i << BB_i.NumVariables     << endl;
   of_i << BB_i.Degree           << endl;
   of_i << BB_i.Probability      << endl;
   of_i << BB_i.NumFactors       << endl;
   of_i << BB_i.SumExps          << endl;
   of_i << BB_i.Sign             << endl;
   of_i << BB_i.DegF             << endl;
   of_i << BB_i.ParFracR         << endl;
   of_i << BB_i.ParFracC         << endl;
   for( i=0; i<BB_i.NumFactors; i++ ) {
      of_i << BB_i.Exps[i] << endl;;
   }
   BB_i.ProbeElm.toString( ZPElm );
   of_i << ZPElm          << endl;
   of_i << BB_i.StaticDeg << endl;
   BB_i.GenElm.toString( ZPElm );
   of_i << ZPElm          << endl;
   for( i=0; i<BB_i.NumVariables-1; i++ ) {
      BB_i.RandA[i].toString( ZPElm );
      of_i << ZPElm << endl;
   }
   for( i=0; i<BB_i.NumVariables-1; i++ ) {
      BB_i.RandC[i].toString( ZPElm );
      of_i << ZPElm << endl;
   }
   for( i=0; i<BB_i.NumVariables; i++ ) {
      BB_i.RandB[i].toString( ZPElm );
      of_i << ZPElm << endl;
   }
   BB_i.Cont.toString( ZPElm );
   of_i << ZPElm << endl;
   of_i << BB_i.pgsem1.dg() << endl;  
   for(i=0; i<=BB_i.pgsem1.dg(); i++ ) {
      BB_i.pgsem1[i].toString( ZPElm );
      of_i << ZPElm << endl;
   }
   // factor images
   for( j=0; j<BB_i.NumFactors; j++ ) {
      of_i << BB_i.FactorImages[j].dg() << endl;
      for(i=0; i<=BB_i.FactorImages[j].dg(); i++ ) {
         (BB_i.FactorImages[j])[i].toString( ZPElm );
         of_i << ZPElm << endl;
      }
   }
   // factor products
   for( j=0; j<BB_i.SumExps; j++ ) {
      of_i << BB_i.FactorProducts[j].dg() << endl;
      for(i=0; i<=BB_i.FactorProducts[j].dg(); i++ ) {
         (BB_i.FactorProducts[j])[i].toString( ZPElm );
         of_i << ZPElm << endl;
      }
   }
   // partial fractions
   for( i=0; i<=BB_i.ParFracC; i++ ) {
      for( j=0; j<=BB_i.ParFracR; j++ ) {
         of_i << BB_i.ParFrac[i][j].dg() << endl;
         for( k=0; k<=BB_i.ParFrac[i][j].dg(); k++ ) {
            (BB_i.ParFrac[i][j])[k].toString( ZPElm );
            of_i << ZPElm << endl;
         } 
      } 
   }

}

template < class K, class KP, class F, class FP, class B, class FB, class H > 
inline ifstream& operator>>(
   ifstream& if_i,
   BlackBoxFactorsHMap<K,KP,F,FP,B,FB,H>& BB_i
   )
{
   int    i,j,k;
   int    dg,type;
   char   ZPElm[MOD_KSTRZ];
   F      Coeff;
   MP_INT MPZero;
   mpz_init_set_ui( &MPZero, 0 );
   if_i >> type;
   if ( type != BBDEFS_FACTOR_HMAP_ID ) {
      OutputErrorExit( BBERROR_FACTORS_HMAP, "Incorrect instance file type" );
   }
   if_i >> BB_i.NumVariables;
   if_i >> BB_i.Degree;
   if_i >> BB_i.Probability;
   if_i >> BB_i.NumFactors;
   if_i >> BB_i.SumExps;
   if_i >> BB_i.Sign;
   if_i >> BB_i.DegF;
   if_i >> BB_i.ParFracR;
   if_i >> BB_i.ParFracC;
   for( i=0; i<BB_i.NumFactors; i++ ) {
      if_i >> BB_i.Exps[i];
   }
   if_i >> ZPElm; BB_i.ProbeElm = F( ZPElm );
   if_i >> BB_i.StaticDeg;
   if_i >> ZPElm; BB_i.GenElm = F( ZPElm );
   for( i=0; i<BB_i.NumVariables-1; i++ ) {
      if_i >> ZPElm; BB_i.RandA[i] = F( ZPElm );
   }
   for( i=0; i<BB_i.NumVariables-1; i++ ) {
      if_i >> ZPElm; BB_i.RandC[i] = F( ZPElm );
   }
   for( i=0; i<BB_i.NumVariables; i++ ) {
      if_i >> ZPElm; BB_i.RandB[i] = F( ZPElm );
   }
   if_i >> ZPElm; BB_i.Cont = F( ZPElm );
   if_i >> dg;
   BB_i.pgsem1 = F( (MP_INT*)&MPZero ); 
   for(i=0; i<=dg; i++ ) {
      if_i >> ZPElm; Coeff = F( ZPElm );
      BB_i.pgsem1 = BB_i.pgsem1 + FP( Coeff, i );
   }
   // factor images
   for( j=0; j<BB_i.NumFactors; j++ ) {
      if_i >> dg;
      BB_i.FactorImages[j] = F( (MP_INT*)&MPZero ); 
      for(i=0; i<=dg; i++ ) {
         if_i >> ZPElm; Coeff = F( ZPElm );
         BB_i.FactorImages[j] = BB_i.FactorImages[j] + FP( Coeff, i );
      }
   }
   // factor products
   for( j=0; j<BB_i.SumExps; j++ ) {
      if_i >> dg;
      BB_i.FactorProducts[j] = F( (MP_INT*)&MPZero ); 
      for(i=0; i<=dg; i++ ) {
         if_i >> ZPElm; Coeff = F( ZPElm );
         BB_i.FactorProducts[j] = BB_i.FactorProducts[j] + FP( Coeff, i );
      }
   }
   // partial fractions
   for( i=0; i<=BB_i.ParFracC; i++ ) {
      for( j=0; j<=BB_i.ParFracR; j++ ) {
         if_i >> dg;
         BB_i.ParFrac[i][j] = F( (MP_INT*)&MPZero );
         for(k=0; k<=dg; k++ ) {
            if_i >> ZPElm; Coeff = F( ZPElm );
            BB_i.ParFrac[i][j] = BB_i.ParFrac[i][j] + FP( Coeff, k );
         }
      }
   }
   mpz_clear( &MPZero );
}

#endif

