
/*-------------------------------------------------------------
  module: bbnumdenhmap.h
  author: Angel L. Diaz

  date: 11/2/96

  description: 
     The class in this module realizes the homorphic map of a 
     numerator/denominator black box. 

  class:
     BlackBoxNumDenHMap

  (c) Copyright 1996 Angel Diaz and Erich Kaltofen, All Rights Reserved
  This code is provided WITHOUT WARRANTY either expressed or implied.
 
-------------------------------------------------------------*/

#ifndef __BBNUMDENHMOD
#define __BBNUMDENHMOD

#include "BlackBox/HomomorphicMaps/bbhmap.h"
#include "BlackBox/Algorithms/bbnumdenaux.h"

/*-------------------------------------------------------------
  class: BlackBoxNumDenHMap
  date: 11/2/96

  description:
     Provides for constructing a homorphic map of a numerator/
     denominator black box object.

-------------------------------------------------------------*/

template < class K, class KP, class F, class FP, class B, class FB, class H > 
class BlackBoxNumDenHMap : public BlackBox< K > {
   protected:
      H* h;
      FB* BB;
      BlackBoxNumDen<K,KP,B>* BBND;
      int ConvertStatus;
      F RandA[BBNUMDENAUX_MAXVARS];  
      F RandB[BBNUMDENAUX_MAXVARS];  
      FP LinForms[BBNUMDENAUX_MAXVARS];  
      int D;
      int E;
      int EHat;                  // bound on deg( g(x1,...,xn) )
      F LC;
      FP FHat;
      FP GHat;
      F GenElm; 
      int StaticDeg;
      F Ind[BBNUMDENAUX_MAXCOEFFS];
      F Reciprocal[BBNUMDENAUX_MAXCOEFFS];
      FP Basis[BBNUMDENAUX_MAXCOEFFS];
      F ValueOfNumDen[2];
   public:
      inline friend ofstream& operator<<(ofstream&, 
         BlackBoxNumDenHMap<K,KP,F,FP,B,FB,H>& );
      inline friend ifstream& operator>>(ifstream&, 
         BlackBoxNumDenHMap<K,KP,F,FP,B,FB,H>& );
     
      BlackBoxNumDenHMap( FB& );
      BlackBoxNumDenHMap( FB&, BlackBoxNumDen<K,KP,B>&, H&);
      ~BlackBoxNumDenHMap( ); 

      inline F* operator () ( F* );
      inline int IsConverted() { return ConvertStatus; }
};

template < class K, class KP, class F, class FP, class B, class FB, class H >
BlackBoxNumDenHMap<K,KP,F,FP,B,FB,H>::BlackBoxNumDenHMap( 
   FB& FB_i
   ) 
{
   BB            = &FB_i;
}

template < class K, class KP, class F, class FP, class B, class FB, class H >
BlackBoxNumDenHMap<K,KP,F,FP,B,FB,H>::BlackBoxNumDenHMap(
   FB& FB_i,
   BlackBoxNumDen<K,KP,B>& BB_i,
   H& h_i
   )
{
   int i,j;
   MP_INT MPZero, MPOne;
   K  KElm;
   F  FElm;
   KP KPoly;
   FP FPoly;
   HPoly< K, KP, F, FP, H > hp( h_i );

   mpz_init_set_ui( &MPZero, 0 );
   mpz_init_set_ui( &MPOne, 1 );
   BB            = &FB_i;
   BBND          = &BB_i;
   h             = &h_i;
   Degree        = BB_i.Deg();
   NumVariables  = BB_i.NumVars();
   Probability   = BB_i.Prob();
   D             = BB_i.GetD();
   E             = BB_i.GetE();
   EHat          = BB_i.GetEHat();

   // reset static interpolation information
   StaticDeg     = -1;
   GenElm        = F( (MP_INT*) &MPOne );
   
   KElm          = BB_i.GetLC();
   FElm          = h_i( KElm );
   if ( h_i.IsConverted() == BBDEFS_FALSE ) {
      ConvertStatus = BBDEFS_FALSE;
      goto XDONE;
   }
   LC = FElm; 

   KPoly          = BB_i.GetFHat();
   FPoly          = hp( KPoly );
   if ( hp.IsConverted() == BBDEFS_FALSE ) {
      ConvertStatus = BBDEFS_FALSE;
      goto XDONE;
   }
   FHat = FPoly;

   KPoly          = BB_i.GetGHat();
   FPoly          = hp( KPoly );
   if ( hp.IsConverted() == BBDEFS_FALSE ) {
      ConvertStatus = BBDEFS_FALSE;
      goto XDONE;
   }
   GHat = FPoly;

   // convert random points ai from a2, ... , an
   // convert random points bi from b2, ... , bn
   for( i=0; i<NumVariables-1; i++ ) {
      KElm = BB_i.GetRandA( i );
      FElm = h_i( KElm );
      if ( h_i.IsConverted() == BBDEFS_FALSE ) {
         ConvertStatus = BBDEFS_FALSE;
         goto XDONE; 
      }
      else
        RandA[i] = FElm;
      KElm = BB_i.GetRandB( i );
      FElm = h_i( KElm );
      if ( h_i.IsConverted() == BBDEFS_FALSE ) {
         ConvertStatus = BBDEFS_FALSE;
         goto XDONE;
      }
      else
        RandB[i] = FElm;
   }

   // convert n linear forms X, a2*X + b2, ..., an*X + bn
   for( i=0; i<NumVariables; i++ ) {
      KPoly          = BB_i.GetLinForms( i );
      FPoly          = hp( KPoly );
      if ( hp.IsConverted() == BBDEFS_FALSE ) {
         ConvertStatus = BBDEFS_FALSE;
         goto XDONE;
      }
      LinForms[i] = FPoly;
   }

   XDONE:
   mpz_clear( &MPZero );
   mpz_clear( &MPOne );
}

template < class K, class KP, class F, class FP, class B, class FB, class H >
BlackBoxNumDenHMap<K,KP,F,FP,B,FB,H>::~BlackBoxNumDenHMap( )
{
}

template < class K, class KP, class F, class FP, class B, class FB, class H >
inline F* BlackBoxNumDenHMap<K,KP,F,FP,B,FB,H>::operator() (
   F* Pnts_i
   )
{
   int i,j,k;
   unsigned int NumYVals, NumYCandidates, MaxCandidates;
   MP_INT MPZero,MPOne;
   mpz_init_set_ui( &MPZero, 0 );
   mpz_init_set_ui( &MPOne, 1 );
   FP TmpFHat,TmpGHat,TmpPoly;
   F  TmpElm; 
   FP FBarArray[BBNUMDENAUX_MAXDEG];
   FP GBarArray[BBNUMDENAUX_MAXDEG];
   F  ConstPnts[BBNUMDENAUX_MAXVARS];   // pi - ai*p1 - bi, i=1,...,n
   FP FullLinForms[BBNUMDENAUX_MAXVARS];  
   F  YArray[BBNUMDENAUX_MAXDEG];
   F  Dep[BBNUMDENAUX_MAXDEG];
   F  I;
   F  Zero,One;

   // compute Y multipliers
   for( i=0; i<NumVariables-1; i++ ) 
      ConstPnts[i] = Pnts_i[i+1] - RandA[i]*Pnts_i[0] - RandB[i]; 

   FBarArray[0]  = FHat;
   GBarArray[0]  = GHat;
   Zero          = F( (MP_INT*)&MPZero );
   One           = F( (MP_INT*)&MPOne );
   YArray[0]     = Zero;
   I             = One;
   NumYVals      = BBDEFS_MAX( D, E ) + 1;
   MaxCandidates = 2*D*E;
   StaticDeg     = -1;
   GenElm        = F( (MP_INT*) &MPOne );
 
#ifdef BBDEFS_NUMDEN_TRC
 BBNUMDENTrace4( Pnts_i, NumVariables, ConstPnts, NumYVals, MaxCandidates );
#endif

   // compute the univariate images of FBar and GBar
   FullLinForms[0] = LinForms[0];
   j = 0; NumYCandidates = 1;
   while ( BBDEFS_TRUE ) {

#ifdef BBDEFS_NUMDEN_TRC
 BBNUMDENTrace8( NumYCandidates );
#endif

      for( i=1; i<NumVariables; i++ ) {
         FullLinForms[i] = LinForms[i] + I*ConstPnts[i-1];
      }
      Pade( BB, D, E, FullLinForms, LC, GenElm, StaticDeg, Ind, Reciprocal,
            Basis, TmpFHat, TmpGHat );
      if ( TmpFHat.dg() >= D && TmpGHat.dg() >= E ) {
         YArray[j+1]    = I;
         FBarArray[j+1] = TmpFHat;
         GBarArray[j+1] = TmpGHat;
         j++;
      }
      NumYCandidates++;
      I.DistinctSelect();
      if ( j == NumYVals ) 
         break;
      if ( NumYCandidates == MaxCandidates )
         OutputErrorExit( BBERROR_NUMDEN, 
                          "Evaluation : Unluky random points" ); 
   }

#ifdef BBDEFS_NUMDEN_TRC
 BBNUMDENTrace5( FBarArray, GBarArray, YArray, j+1 );
#endif 

   StaticDeg = -1;
   // interpolate FBar 
   ValueOfNumDen[0] = Zero;
   TmpElm = One;
   for( i=0; i<=j; i++ ) {
      for( k=0; k<=j; k++ ) {
         Dep[k] = (FBarArray[k])[i];
      }
      UniDenseInterp( YArray, Dep, j, StaticDeg, Reciprocal,
                      Basis, TmpPoly ); 
      ValueOfNumDen[0] = ValueOfNumDen[0] + TmpElm * TmpPoly.eval( &One );
      TmpElm = TmpElm * Pnts_i[0];
   }  

   // interpolate GBar
   ValueOfNumDen[1] = Zero;
   TmpElm = One;
   for( i=0; i<=j; i++ ) {
      for( k=0; k<=j; k++ ) {
         Dep[k] = (GBarArray[k])[i];
      }
      UniDenseInterp( YArray, Dep, j, StaticDeg, Reciprocal,
                      Basis, TmpPoly ); 
      ValueOfNumDen[1] = ValueOfNumDen[1] + TmpElm * TmpPoly.eval( &One );
      TmpElm = TmpElm * Pnts_i[0];
   }  
 
#ifdef BBDEFS_NUMDEN_TRC
 BBNUMDENTrace6( ValueOfNumDen[0], ValueOfNumDen[1] );
#endif

   mpz_clear( &MPZero );
   mpz_clear( &MPOne );
 
   return( ValueOfNumDen );
}

template < class K, class KP, class F, class FP, class B, class FB, class H > 
inline ofstream& operator<<(
   ofstream& of_i,
   BlackBoxNumDenHMap<K,KP,F,FP,B,FB,H>& BB_i
   )
{
}

template < class K, class KP, class F, class FP, class B, class FB, class H > 
inline ifstream& operator>>(
   ifstream& if_i,
   BlackBoxNumDenHMap<K,KP,F,FP,B,FB,H>& BB_i
   )
{
}

#endif

