
/*-------------------------------------------------------------
  module: shoup.h

  author: Angel L. Diaz
  date: 11/24/96

  Realizes a black box compliant univariate polynomial and
  field arithmetic plug-in to v. shoups package.
-------------------------------------------------------------*/

#ifndef _SHOUP_PLUGIN
#define _SHOUP_PLUGIN

#define SHOUP_SHMaxRand 2147483647

#include "PlugIns/Shoup/ZZ_pX.h"
#include "PlugIns/Shoup/ZZ_pXFactoring.h"
#include "BlackBox/Aux/bbdefs.h"
#include "PlugIns/gmp.h"

class ShoupZP;         
class ShoupZPX; 
class ShoupZPXT;

MP_INT SHOUP_MPCard; 
ZZ     SHOUP_SHCard;
int    SHOUP_TruncDeg;

ZZ     ShoupBoundMPInt( const MP_INT*, MP_INT* );
inline ZZ ShoupBinarySearch( const MP_INT* );
inline ZZ ConvertMPToZZ( const MP_INT* );
int    ZZToString( const ZZ&, int, char* );
void   StringToZZ( const char*, int&, ZZ& );
int    VectorZZpToString( const vector(ZZ_p)&, int, char* );
void   StringToVectorZZp( const char*, int&, vector(ZZ_p)& );
void   ShoupInitEnv( MP_INT* );
void   ShoupCleanUpEnv( ); 

/*-------------------------------------------------------------
  Description: Auxiliary functions and procedures

  date: 8/15/96

  Utilized by the black box compliant field and polynomial 
  arithmetic plug-in to v shoup, as well as black box applications.
-------------------------------------------------------------*/

inline ZZ ShoupBoundMPInt( 
   const MP_INT *Key_i, 
   MP_INT *Bound_o 
   ) 
{
   ZZ Bound(INIT_VAL, 1);
   mpz_set_ui( Bound_o, 1 );

   while ( mpz_cmp( Key_i, Bound_o ) > 0 ) {
      mpz_mul_ui( Bound_o, Bound_o, 2);
      mul( Bound, Bound, 2 );
   }

   return Bound;
}

inline ZZ ShoupBinarySearch(
   const MP_INT* Key
   ) 
{
   MP_INT MPPivot; 
   mpz_init_set_ui( &MPPivot, 0 );
   ZZ SHPivot(INIT_VAL, 0);
   ZZ SHOne(INIT_VAL, 1);
   ZZ SHTwo(INIT_VAL, 2);

   MP_INT MPRightEnd; 
   mpz_init( &MPRightEnd );
   ZZ SHRightEnd = ShoupBoundMPInt( Key, &MPRightEnd );

   MP_INT MPLeftEnd; 
   mpz_init_set_ui( &MPLeftEnd, 1);
   ZZ SHLeftEnd(INIT_VAL, 1);

   while ( mpz_cmp( &MPRightEnd, &MPLeftEnd ) >= 0 ) {
      mpz_add( &MPPivot, &MPLeftEnd, &MPRightEnd ); 
      add( SHPivot, SHLeftEnd, SHRightEnd );

      mpz_div_ui( &MPPivot, &MPPivot, 2 );  
      div( SHPivot, SHPivot, SHTwo );

      if ( mpz_cmp( Key, &MPPivot ) == 0 ) break;
      if ( mpz_cmp( Key, &MPPivot ) < 0 ) { 
         mpz_sub_ui( &MPRightEnd, &MPPivot, 1 ); 
         sub( SHRightEnd, SHPivot, SHOne ); 
      }
      else { 
         mpz_add_ui( &MPLeftEnd, &MPPivot, 1 ); 
         add( SHLeftEnd, SHPivot, SHOne );
      }
   }

   mpz_clear( &MPRightEnd );
   mpz_clear( &MPPivot );
   mpz_clear( &MPLeftEnd );

   return SHPivot;
}

inline ZZ ConvertMPToZZ(
   const MP_INT *MPInt_i
   ) 
{
   ZZ SHInt(INIT_VAL, 0);
   MP_INT MPInt;
   int IsPositive;

   mpz_init_set( &MPInt, MPInt_i );
   if ( mpz_cmp_si( &MPInt, 0 ) != 0 ) {

      IsPositive = Positive( &MPInt );
      if ( IsPositive == BBDEFS_FALSE ) 
         mpz_abs( &MPInt, &MPInt);

      SHInt = ShoupBinarySearch( &MPInt ); 

      if ( IsPositive == BBDEFS_FALSE ) 
         mul( SHInt, SHInt, -1 );
   }
   mpz_clear( &MPInt ); 
   return SHInt;
}

struct astack {
   long top;
   long alloc;
   long *elts;

   astack() { top = -1; alloc = 0; elts = 0; }
   ~astack() { }

   long pop() { return elts[top--]; }
   long empty() { return (top == -1); }
   void push(long x);
};

void astack::push(long x)
{
   if (alloc == 0) {
      alloc = 100;
      elts = (long *) malloc(alloc * sizeof(long));
   }

   top++;

   if (top + 1 > alloc) {
      alloc = 2*alloc;
      elts = (long *) realloc(elts, alloc * sizeof(long));
   }

   if (!elts) {
      Error("out of space in ZZ output");
   }

   elts[top] = x;
}

int ZZToString(
   const ZZ& a_i,
   int i_i, 
   char* Pntr_m
   )
{
   static ZZ b;
   static astack S;
   long r;
   long k;

   b = a_i;

   k = sign(b);

   if (k == 0) {
      Pntr_m[i_i] = '0'; i_i++; 
      Pntr_m[i_i] = '\0';
      return i_i; // location of \0
    }
    
    if (k < 0) {
        Pntr_m[i_i] = '-'; i_i++;
      negate(b, b);
   }

   do {
      r = DivRem(b, b, 10);
      S.push(r);
   } while (!IsZero(b));

   while (!S.empty()) {
      r = S.pop();
      Pntr_m[i_i] = ((char) (r + '0')); i_i++;
   }

    Pntr_m[i_i] = '\0';
    return i_i; // location of \0
}

void StringToZZ( 
   const char* Pntr_i, 
   int& i_m,
   ZZ&  a
   )
{
   char c;
   int sign;
    
   a << 0;
    
   c = 0; 
   c = Pntr_i[i_m]; i_m++;
   if (c == '-') {
       sign = -1;
       c = 0; 
       c = Pntr_i[i_m]; i_m++;
   }
   else
       sign = 1;
    
   if (c < '0' || c > '9') Error(" bad ZZ input");
    
   while (c >= '0' && c <= '9') {
       mul(a, a, 10);
       add(a, a, c - '0');
       c = 0; c = Pntr_i[i_m]; i_m++;
   }
 
   if ( c == ']' )
      i_m--;
   else
   if ( c != '\0' && c != ' ' ) 
      Error(" bad ZZ input");
   
   if (sign == -1)
       negate(a, a);
    
   return;
}

int VectorZZpToString( 
   const vector(ZZ_p)& a,
   int i_i, 
   char* Pntr_m 
   )
{
   int i,j, n; 
             
   Pntr_m[i_i] = '['; i_i++;
   n = a.length(); 

   for (i = 0,j=1; i < n; i++,j++) { 
      i_i = ZZToString( rep( a[i] ), i_i, Pntr_m );
      if (i < n-1) { Pntr_m[i_i] = ' '; i_i++; }
   }  

   Pntr_m[i_i] = ']'; i_i++;
   Pntr_m[i_i] = '\0' ;

   return i_i; // location of \0
}

void StringToVectorZZp( 
   const char* Pntr_i, 
   int& i_m,
   vector(ZZ_p)& a
   )
{
   char c;  
   int  n;   
   ZZ   TmpZZ;

   c = 0; c = Pntr_i[i_m]; i_m++;

   if (c != '[') Error("bad Vector input\n"); 

   n = 0;  
   a.SetLength(0);  

   c = Pntr_i[i_m]; i_m++;

   while (c != ']') {   
      i_m--;
      n++; 
      a.SetLength(n);
      StringToZZ( Pntr_i, i_m, TmpZZ );
      a[n-1] << TmpZZ; 
      c = Pntr_i[i_m]; i_m++;
   }  
   i_m++; // location of \0
   return;    
}

void ShoupInitEnv(
   MP_INT* MPCard_i
   ) 
{ 
  mpz_init_set( &SHOUP_MPCard, MPCard_i ); 
  SHOUP_SHCard =  ShoupBinarySearch( &SHOUP_MPCard );

  ZZ_pInit( SHOUP_SHCard ); 
}

void ShoupCleanUpEnv( ) { 
}

/*-------------------------------------------------------------
  class: ShoupZP

  date: 8/6/96

  Black box compliant field arithmetic plug-in to v. shoup.
-------------------------------------------------------------*/

class  ShoupZP {
   private:
      ZZ_p ShoupZPElm; 
   public:
      friend ShoupZPX;
      friend ShoupZPXT;
      inline friend ostream& operator<<( ostream&, const ShoupZP& );

      ShoupZP();
      ShoupZP( const long& );
      ShoupZP( const ZZ& );
      ShoupZP( const ZZ_p& );
      ShoupZP( const MP_INT* );
      ShoupZP( const char* );
      ShoupZP( const ShoupZP& );
      ~ShoupZP();
  
      inline ShoupZP& operator= ( const ShoupZP&  );
      inline friend int operator==( const ShoupZP&, const ShoupZP& );
      inline friend int operator!=( const ShoupZP&, const ShoupZP& );

      inline ShoupZP operator*( const ShoupZP& );
      inline ShoupZP operator/( const ShoupZP& );
      inline ShoupZP operator+( const ShoupZP& );
      inline ShoupZP operator-( const ShoupZP& );

      inline ShoupZP DistinctSelect( );
      inline MP_INT card( )               { return SHOUP_MPCard; }
      inline ShoupZP Exp( unsigned int );
      inline ShoupZP Denom( );
      inline ShoupZP LCM( ShoupZP*, unsigned int );
      inline void toString( char* ); 

      class RandomIterator {
         public:
            RandomIterator(const unsigned int, const MP_INT*);
            ~RandomIterator()        { mpz_clear(&MPCard); }
            ShoupZP operator()();
         private:
            long RandStateArray[32];
            unsigned int RandSeed;
            MP_INT MPCard;
            ZZ SHCard;
      };

};

inline ostream& operator<<( 
   ostream& os, 
   const ShoupZP& FielElm_i
  )
{
   os << FielElm_i.ShoupZPElm;
   return os;
}

ShoupZP::ShoupZP( ) { 
}

ShoupZP::ShoupZP( 
   const long& FielElm_i 
   )   
{ 
   ShoupZPElm << FielElm_i; 
}

ShoupZP::ShoupZP( 
   const ZZ& FielElm_i ) 
{ 
   ShoupZPElm << FielElm_i; 
}

ShoupZP::ShoupZP( 
   const ZZ_p& FielElm_i 
   )   
{ 
   ShoupZPElm = FielElm_i; 
} 

ShoupZP::ShoupZP( const MP_INT *MPInt_i ) {
    ShoupZPElm << ConvertMPToZZ( MPInt_i );
}

ShoupZP::ShoupZP( const char* Pntr_i ) {
   ZZ Elm;
   int loc = 0;
   StringToZZ( Pntr_i, loc, Elm );
   ShoupZPElm << Elm;
}

ShoupZP::ShoupZP(
   const ShoupZP& FielElm_i
   )    
{ 
   ShoupZPElm = FielElm_i.ShoupZPElm; 
}

ShoupZP::~ShoupZP( ) { 
}

inline ShoupZP& ShoupZP::operator= ( 
   const ShoupZP& FielElm_i
   )
{ 
   ShoupZPElm = FielElm_i.ShoupZPElm; 
   return *this; 
} 

inline int operator==(
   const ShoupZP& FielElmA_i, 
   const ShoupZP& FielElmB_i
   )
{ 
   return( FielElmA_i.ShoupZPElm == FielElmB_i.ShoupZPElm ); 
}

inline int operator!=(
   const ShoupZP& FielElmA_i, 
   const ShoupZP& FielElmB_i
   )
{ 
   return( FielElmA_i.ShoupZPElm != FielElmB_i.ShoupZPElm ); 
}

inline ShoupZP ShoupZP::operator*(
   const ShoupZP& FielElm_i
   )
{ 
   ShoupZP K ; 
   mul( K.ShoupZPElm, this->ShoupZPElm, FielElm_i.ShoupZPElm ); 
   return K; 
}

inline ShoupZP ShoupZP::operator/(
   const ShoupZP& FielElm_i
   )
{  
   ShoupZP K; 
   inv( K.ShoupZPElm, FielElm_i.ShoupZPElm ); 
   if ( *this == (ShoupZP) 1 ) 
      return K;
   return ( (*this) * K ); 
}

inline ShoupZP ShoupZP::operator+( 
   const ShoupZP& FielElm_i 
   )
{ 
   ShoupZP K; 
   add( K.ShoupZPElm, this->ShoupZPElm, FielElm_i.ShoupZPElm ); 
   return K; 
}

inline ShoupZP ShoupZP::operator-(
   const ShoupZP& FielElm_i
   )
{ 
   ShoupZP K; 
   sub( K.ShoupZPElm, this->ShoupZPElm, FielElm_i.ShoupZPElm ); 
   return K; 
}

inline ShoupZP ShoupZP::DistinctSelect( ) {
   ShoupZP K;
   K = *this;

   add( this->ShoupZPElm, this->ShoupZPElm, 1); 
   return K;
}

inline ShoupZP ShoupZP::Exp( 
   unsigned int n_i
   ) 
{
   unsigned int i;
   int Sign;
   ZZ_p Prod;
   ZZ_p InvProd;
 
   if ( n_i == 0 ) {
      Prod << 1;
      return ShoupZP( Prod );
   }
   else
      Prod = ShoupZPElm;

    if ( n_i < 0 ) {
       Sign = -1;
       n_i = n_i * -1;
    }
    else
       Sign = 1;

   for(i=1;i<n_i;i++) 
      mul( Prod, Prod, ShoupZPElm );

   if ( Sign == 1 )
       return ShoupZP( Prod );
    else {
       inv( InvProd, Prod ); 
       return ShoupZP( InvProd );
    }

}

inline ShoupZP ShoupZP::Denom( ) {
   return ShoupZP( (long)1 );
}

inline ShoupZP ShoupZP::LCM(
   ShoupZP* ShoupZPElmArray_i, 
   unsigned int NumElms_i 
   ) 
{
   return ShoupZP( (long)1);
}

inline void ShoupZP::toString(
   char* Pntr_m
   ) 
{ 
   ZZToString( rep( ShoupZPElm ), 0, Pntr_m);
}

ShoupZP::RandomIterator::RandomIterator(
   const unsigned int Seed_i, 
   const MP_INT* Card_i
   ) : RandSeed(Seed_i) 
{
   if ( mpz_cmp_ui( Card_i, 0 ) < 0 ) 
      OutputErrorExit( BBERROR_SHOUP, 
         "ShoupZP::RandomIterator : input card < 0" ); 

   mpz_init_set( &MPCard, Card_i );
   SHCard = ConvertMPToZZ( &MPCard );

   for(int i=0;i<32;i++) 
      RandStateArray[i] = (long)i;
   initstate( Seed_i, (char*) RandStateArray, sizeof(RandStateArray) );
}

ShoupZP ShoupZP::RandomIterator::operator()( ) 
{
   long r;
   ZZ c,x;
   char *PrevState;

   x << 0; 
   c << 1;

   PrevState = (char*)setstate((char*)RandStateArray);

   while ( c < SHCard ) {   
      r = random()-1;           // random long [0,2147483647)
      mul( x, x, SHOUP_SHMaxRand);
      add( x, x, r );
      mul( c, c, SHOUP_SHMaxRand);
   }
  
   rem(x, x, SHCard);
   setstate(PrevState);

   return( ShoupZP( x ) );
}

/*-------------------------------------------------------------
  class: ShoupZPX

  date: 8/6/96

  Black box compliant univariate polynomial arithmetic plug-in 
  to v. shoup.
-------------------------------------------------------------*/

class ShoupZPX {
   private:
      ZZ_pX ElmZPX;
   public: 
      friend ostream& operator<<( ostream&, const ShoupZPX& );
      friend ShoupZPXT;

      ShoupZPX( );
      ShoupZPX( const long& );
      ShoupZPX( const ZZ_pX& );
      ShoupZPX( const ZZ_p& );
      ShoupZPX( const ShoupZP& );
      ShoupZPX( const ShoupZPX& );
      ShoupZPX( const char* );
      ShoupZPX( const ShoupZP&, unsigned int ); 
      ~ShoupZPX( );

      inline ShoupZPX& operator= (const ShoupZPX& );
      inline friend int operator==(const ShoupZPX&, const ShoupZPX& );
      inline friend int operator!=(const ShoupZPX&, const ShoupZPX& );

      inline ShoupZPX operator*( const ShoupZPX& );
      inline ShoupZPX operator/( const ShoupZPX& );
      inline ShoupZPX operator+( const ShoupZPX& );
      inline ShoupZPX operator-( const ShoupZPX& );

      inline ShoupZP operator[] ( unsigned int i );
      inline unsigned int dg( );
      inline ShoupZP eval( const ShoupZP* );
      inline ShoupZP lcoeff( ); 
      inline void pgcd(const ShoupZPX&, const ShoupZPX&, ShoupZPX& ); 
      inline void prem(const ShoupZPX&, const ShoupZPX&, ShoupZPX& );
      inline void pquorem(const ShoupZPX&, const ShoupZPX&, ShoupZPX&, 
                          ShoupZPX&);
      inline void peea( const ShoupZPX&, const ShoupZPX&, ShoupZPX&,
                        ShoupZPX&, ShoupZPX& );
      inline void factors( const ShoupZPX&, int&, ShoupZP&, unsigned int&,
                           unsigned int*, ShoupZPX* );
      inline ShoupZPX Diff( );

      inline void toString( char* ); 
};

ostream& operator<<(
   ostream& os, 
   const ShoupZPX& ShoupZPX_i
   )
{
   os << ShoupZPX_i.ElmZPX;
   return os;
}

ShoupZPX::ShoupZPX( ) {
}

ShoupZPX::ShoupZPX( 
   const long& ElmZPX_i 
   )       
{ 
   ElmZPX << ElmZPX_i;
}

ShoupZPX::ShoupZPX(
   const ZZ_pX& ElmZPX_i
   )      
{ 
   ElmZPX = ElmZPX_i;
}

ShoupZPX::ShoupZPX(
   const ZZ_p& ShoupZPElm_i
   )       
{ 
   ElmZPX << ShoupZPElm_i;           
}

ShoupZPX::ShoupZPX(
   const ShoupZP& ShoupZPElm_i
   )        
{ 
   ElmZPX << ShoupZPElm_i.ShoupZPElm;
}

ShoupZPX::ShoupZPX(
   const ShoupZPX& ShoupZPX_i
   )       
{ 
   ElmZPX = ShoupZPX_i.ElmZPX;  
}

ShoupZPX::ShoupZPX( const char* Pntr_i ) {
   vector(ZZ_p) Elm;
   int loc = 0;
   StringToVectorZZp( Pntr_i, loc, Elm );
   ElmZPX << Elm;
}

ShoupZPX::ShoupZPX(
   const ShoupZP& ShoupZPElm_i, 
   unsigned int Deg_i
   ) 
{ 
   SetCoeff( ElmZPX, Deg_i , ShoupZPElm_i.ShoupZPElm ); 
}

ShoupZPX::~ShoupZPX( ) { 
}

inline ShoupZPX& ShoupZPX::operator= (
   const ShoupZPX& ShoupZPX_i
   )
{ 
   ElmZPX = ShoupZPX_i.ElmZPX; 
   return *this; 
}

inline int operator==(
   const ShoupZPX& ShoupZPXA_i, 
   const ShoupZPX& ShoupZPXB_i
   )
{ 
   return( ShoupZPXA_i.ElmZPX == ShoupZPXB_i.ElmZPX ); 
}

inline int operator!=(
   const ShoupZPX& ShoupZPXA_i, 
   const ShoupZPX& ShoupZPXB_i
   )
{ 
   return( ShoupZPXA_i.ElmZPX != ShoupZPXB_i.ElmZPX ); 
}

inline ShoupZPX ShoupZPX::operator*(
   const ShoupZPX& ShoupZPXA_i
   )
{ 
   ShoupZPX P; 
   mul( P.ElmZPX, this->ElmZPX, ShoupZPXA_i.ElmZPX ); 
   return P; 
}

inline ShoupZPX ShoupZPX::operator/(
   const ShoupZPX& ShoupZPXA_i
   )
{ 
   ShoupZPX q; 
   divide( q.ElmZPX, this->ElmZPX, ShoupZPXA_i.ElmZPX );
   return q; 
}

inline ShoupZPX ShoupZPX::operator+(
   const ShoupZPX& ShoupZPXA_i
   )
{ 
   ShoupZPX P; 
   add( P.ElmZPX, this->ElmZPX, ShoupZPXA_i.ElmZPX ); 
   return P; 
}

inline ShoupZPX ShoupZPX::operator-( 
   const ShoupZPX& ShoupZPXA_i
   )
{ 
   ShoupZPX P; 
   sub( P.ElmZPX, this->ElmZPX, ShoupZPXA_i.ElmZPX ); 
   return P; 
}

inline ShoupZP ShoupZPX::operator[] (
   unsigned int Deg_i
   ) 
{ 
   ShoupZP K; 
   GetCoeff( K.ShoupZPElm, ElmZPX, Deg_i); 
   return K;
}

inline unsigned int ShoupZPX::dg( ) { 
   return deg( ElmZPX ); 
}

inline ShoupZP ShoupZPX::eval(
   const ShoupZP pnts_i[]
   )
{
   int i;
   int d = deg( ElmZPX );
   ShoupZP res;
   ShoupZP tmp1 = 1;
   ZZ_p tmp2;

   GetCoeff( res.ShoupZPElm, ElmZPX, 0 );
   for(i=1;i<=d;i++) { 
      tmp1 = tmp1 * pnts_i[0]; 
      GetCoeff( tmp2, ElmZPX, i);
      res = res + tmp1 * tmp2; 
   }

   return( res );
}

inline ShoupZP ShoupZPX::lcoeff( ) 
{ 
   ShoupZP K; 
   GetCoeff( K.ShoupZPElm, ElmZPX, deg( ElmZPX ) ); 
   return K;
}

inline void ShoupZPX::pgcd(
   const ShoupZPX& ShoupZPXA_i, 
   const ShoupZPX& ShoupZPXB_i,
   ShoupZPX& ShoupZPX_o
   ) 
{ 
   GCD( ShoupZPX_o.ElmZPX, ShoupZPXA_i.ElmZPX, ShoupZPXB_i.ElmZPX ); 
}

inline void ShoupZPX::pquorem( 
   const ShoupZPX& ShoupZPXA_i, 
   const ShoupZPX& ShoupZPXB_i,
   ShoupZPX& ShoupZPXQ_o,
   ShoupZPX& ShoupZPXR_o
   )
{
   DivRem( ShoupZPXQ_o.ElmZPX, ShoupZPXR_o.ElmZPX, ShoupZPXA_i.ElmZPX, 
           ShoupZPXB_i.ElmZPX); 

}

inline void ShoupZPX::prem(
   const ShoupZPX& ShoupZPXA_i, 
   const ShoupZPX& ShoupZPXB_i,
   ShoupZPX& ShoupZPXR_o
   )
{
   rem( ShoupZPXR_o.ElmZPX, ShoupZPXA_i.ElmZPX, ShoupZPXB_i.ElmZPX); 
}

inline void ShoupZPX::peea(
   const ShoupZPX& ShoupZPXA_i,
   const ShoupZPX& ShoupZPXB_i,
   ShoupZPX& ShoupZPXG_m,
   ShoupZPX& ShoupZPXS_m,
   ShoupZPX& ShoupZPXT_m
   )
{
   // d = gcd(a,b), a s + b t = d
   XGCD( ShoupZPXG_m.ElmZPX, ShoupZPXS_m.ElmZPX, ShoupZPXT_m.ElmZPX,
         ShoupZPXA_i.ElmZPX, ShoupZPXB_i.ElmZPX );
}

inline void ShoupZPX::factors(
   const ShoupZPX& ShoupZPX_i,
   int& Sign_o,
   ShoupZP& Cont_o,
   unsigned int& NumFactors_o,
   unsigned int *Exps_o,
   ShoupZPX *Factors_o
   )
{
   int i;
   ZZ_p  LC, LCInv;
   ZZ_pX MonicPoly;
   vector(ZZ_pXIntPair) Facs;

   GetCoeff( LC, ShoupZPX_i.ElmZPX, deg( ShoupZPX_i.ElmZPX ) ); 
   inv( LCInv, LC); 
   mul( MonicPoly, ShoupZPX_i.ElmZPX, LCInv );
   berlekamp( Facs, MonicPoly ); 

   Sign_o            = 1;
   Cont_o.ShoupZPElm = LC;
   NumFactors_o      = Facs.length();
   for(i=0; i<NumFactors_o; i++) {
      Factors_o[i] = Facs[i].g;
      Exps_o[i]    = Facs[i].k;
   }
}

inline ShoupZPX ShoupZPX::Diff( ) 
{
   ZZ_pX P;

   diff( P, ElmZPX );

   return ( ShoupZPX( P ) );
}

inline void ShoupZPX::toString( 
   char* Pntr_m
   ) 
{
   VectorZZpToString( ElmZPX.rep, 0, Pntr_m );
}

/*-------------------------------------------------------------
  class: ShoupZPXT

  date: 10/7/96
-------------------------------------------------------------*/

class ShoupZPXT {
   private:
      ZZ_pX ElmZPX;
   public:
      friend ostream& operator<<( ostream&, const ShoupZPXT& );

      ShoupZPXT( ); 
      ShoupZPXT( const ShoupZP& );
      ShoupZPXT( const ShoupZPX& );
      ~ShoupZPXT( );

};

inline void SetTruncDeg( 
   const int d_i 
   )
{
   SHOUP_TruncDeg = d_i;
}

ostream& operator<<(
   ostream& os,
   const ShoupZPXT& ShoupZPXT_i
   )
{
   os << ShoupZPXT_i.ElmZPX;
   return os;
}

ShoupZPXT::ShoupZPXT( ) { }

ShoupZPXT::ShoupZPXT( 
   const ShoupZP& ShoupZPElm_i
   ) 
{ 
   ElmZPX << ShoupZPElm_i.ShoupZPElm;
   //trunc( ElmZPX, ElmZPX, TruncDefg);
}

ShoupZPXT::ShoupZPXT( 
   const ShoupZPX& ShoupZPXElm_i 
   ) 
{
   ElmZPX = ShoupZPXElm_i.ElmZPX;
}

ShoupZPXT::~ShoupZPXT( ) { }

#endif
